// vim: autoindent tabstop=8 shiftwidth=2 expandtab softtabstop=2
#include <stdio.h>
#include <string.h>
#include <limits.h>     // realpath(3)
#include <stdlib.h>     // realpath(3)

// Handle relative pathnames - erco@seriss.com
// Ref: realpath(1)
// TODO: Remove all duplicate /'s and /./'s after calling realpath().
//       (realpath() does this, but only to real paths, not "fictitious" paths)

// Use this function to build a string without overrunning its buffer.
// Safely copy 's' into 'to', keeping count of remaining chars.
//
// On return:
//     > 'to' is adjusted to point to the new terminating NULL
//     > 'tolen' is decreased by the number of characters added (if any).
//     > Unless tolen was 0 on entry, 'to' is always left terminated.
//
// Returns 0 on success (as described above),
// or -1 if an overrun was detected, leaving 'to' truncated + terminated.
//
// Expected use:
//     // Allocate string
//     int outlen = 80;
//     char *out = new char[outlen];
//     // Construct string
//     int fail = 0;
//     char *op = out;
//     fail |= safe_strcpy(op, outlen, "some text..");
//     fail |= safe_strcpy(op, outlen, "more text..");
//     printf("out is: '%s'\n", out);
//     if ( fail == -1 ) { printf("(out was truncated)\n"); }
//
static int safe_strcpy(char *&to, int &tolen, const char *s) {
  if ( tolen == 0 ) return -1;
  while ( tolen > 1 ) {
    if ( (*to = *s) == 0 ) break;
    ++to; ++s; --tolen;
  }
  *to = '\0';
  return *s == 0 ? 0 : -1;
}


// Return index of first path component different between paths a and b
static int find_path_diff(const char *a, const char *b) {
  int i = 0;
  int slash = 0;
  while ( *a && *b && (*a == *b) ) {  // loop until diff found or end of a|b
    if ( *a == '/' ) slash = i+1;
    ++i; a++; b++;
  }
  if ( *a==0 && (*b==0 || *b=='/') ) return i;  // end of a and b (or b is /)
  if ( *b==0 && (*a==0 || *a=='/') ) return i;
  return slash;
}

int                                      // O - 0 if no change, 1 if changed, -1 if overrun
fl_filename_relative(char       *to,     // O - Relative filename
                     int        tolen,   // I - Size of "to" buffer
                     const char *from,   // I - Absolute filename
                     const char *base)   // I - Find path relative to this path
{
  int fail = 0;
  const char *from_realpath, *base_realpath;

  // Canonicalize input paths --  removes redundant /../'s and /./'s
  if ( !(from_realpath = realpath(from, 0)) ) from_realpath = strdup(from);
  if ( !(base_realpath = realpath(base, 0)) ) base_realpath = strdup(base);
  // Find index of first complete component where paths differ..
  int index = find_path_diff(from_realpath, base_realpath);
  const char *from_diff = from_realpath + index;
  const char *base_diff = base_realpath + index;
  if ( *from_diff == '/' ) from_diff++;
  if ( *base_diff == '/' ) base_diff++;
  //DEBUG printf("    FROMDIFF: '%s'\n", from_diff);
  //DEBUG printf("    BASEDIFF: '%s'\n", base_diff);
  if ( *base_diff ) {                            // [1] leftovers in base?
    //DEBUG printf("[1]\n");
    fail |= safe_strcpy(to, tolen, "..");        // [1] go up at least one dir
    while ( *base_diff ) {                       // [1] search for /'s in base_diff,
      if ( *base_diff++ == '/' ) {               // [1] ..and for every remaining / in base_diff,
        if ( *base_diff ) {                      // [1] ..as long as it's not an extra trailing slash,
          fail |= safe_strcpy(to, tolen, "/.."); // [1] .. go up one dir
        }
      }
    }
    if ( *from_diff ) {                          // [2] Also leftovers in from_diff?
      //DEBUG printf("[2]\n");
      fail |= safe_strcpy(to, tolen, "/");       // [2] add slash
      fail |= safe_strcpy(to, tolen, from_diff); // [2] and rest of from_diff
    }
  } else if ( *from_diff ) {                     // [3] leftovers in from only? then we're in base dir
    //DEBUG printf("[3]\n");
    fail |= safe_strcpy(to, tolen, from_diff);   // [3] just add on from_diff for relative path
  } else {
    //DEBUG printf("[4]\n");
    fail |= safe_strcpy(to, tolen, ".");         // Same from + base? "."
  }
  // Dump canon paths
  free((void*)from_realpath);
  free((void*)base_realpath);
  return fail ? -1 : 0;
}

class RegTest {
public:
    const char *from;
    const char *base;
    const char *expect;
    const char *comment;
};

int main() {
  char buf[256];
  RegTest test[] = { 
   // From              Base           Expect             Comments
   { "/tmp",            "/tmp/a/b/c/", "../../..",        "[1] Leftovers in Base Extra Trailing Base Slash" },
   { "/tmp",            "/tmp/a/b/c",  "../../..",        "[1] Leftovers in Base" },
   { "/tmp",            "/tmp/a/b/",   "../..",           "[1] Leftovers in Base" },
   { "/tmp",            "/tmp/a/b",    "../..",           "[1] Leftovers in Base" },
   { "/tmp",            "/tmp/a/",     "..",              "[1] Leftovers in Base" },
   { "/tmp",            "/tmp/a",      "..",              "[1] Leftovers in Base" },
   { "/tmp",            "/tmp/",       ".",               "[4] Same With Extra Trailing Base Slash" },
   { "/tmp",            "/tmp",        ".",               "[4] Leftovers in Base" },
   { "/tmp",            "/",           "tmp",             "[3] Leftovers in Base" },

   { "/tmp/a",          "/tmp/a/b/c/", "../..",           "[1] Leftovers in Base Extra Trailing Base Slash" },
   { "/tmp/a",          "/tmp/a/b/c",  "../..",           "[1] Leftovers in Base" },
   { "/tmp/a",          "/tmp/a/b/",   "..",              "[1] Leftovers in Base" },
   { "/tmp/a",          "/tmp/a/b",    "..",              "[1] Leftovers in Base" },
   { "/tmp/a",          "/tmp/a/",     ".",               "[4] Same with Extra Trailing Base Slash" },
   { "/tmp/a",          "/tmp/a",      ".",               "[4] Same" },
   { "/tmp/a",          "/tmp/",       "a",               "[3] Leftovers in From Only" },
   { "/tmp/a",          "/tmp",        "a",               "[3] Leftovers in From Only" },
   { "/tmp/a",          "/",           "tmp/a",           "[3] Leftovers in From Only" },

   { "/tmp/z",          "/tmp/a/b/c/", "../../../z",      "[2] Leftovers in Base+From Extra Trailing Base" },
   { "/tmp/z",          "/tmp/a/b/c",  "../../../z",      "[2] Leftovers in Base+From" },
   { "/tmp/z",          "/tmp/a/b/",   "../../z",         "[2] Leftovers in Base+From" },
   { "/tmp/z",          "/tmp/a/b",    "../../z",         "[2] Leftovers in Base+From" },
   { "/tmp/z",          "/tmp/a/",     "../z",            "[2] Leftovers in Base+From" },
   { "/tmp/z",          "/tmp/a",      "../z",            "[2] Leftovers in Base+From" },
   { "/tmp/z",          "/tmp/",       "z",               "[3] Leftovers in From Only" },
   { "/tmp/z",          "/tmp",        "z",               "[3] Leftovers in From Only" },
   { "/tmp/z",          "/",           "tmp/z",           "[3] Leftovers in From Only" },

   { "/tmp/xyz",        "/tmp/a/b/c/", "../../../xyz",    "[2] Leftovers in Base+From Extra Trailing Base" },
   { "/tmp/xyz",        "/tmp/a/b/c",  "../../../xyz",    "[2] Leftovers in Base+From" },
   { "/tmp/xyz",        "/tmp/a/b/",   "../../xyz",       "[2] Leftovers in Base+From" },
   { "/tmp/xyz",        "/tmp/a/b",    "../../xyz",       "[2] Leftovers in Base+From" },
   { "/tmp/xyz",        "/tmp/a/",     "../xyz",          "[2] Leftovers in Base+From" },
   { "/tmp/xyz",        "/tmp/a",      "../xyz",          "[2] Leftovers in Base+From" },
   { "/tmp/xyz",        "/tmp/",       "xyz",             "[2] Leftovers in Base+From" },
   { "/tmp/xyz",        "/tmp",        "xyz",             "[2] Leftovers in Base+From" },
   { "/tmp/xyz",        "/",           "tmp/xyz",         "[2] Leftovers in Base+From" },

   { "/tmp/a",          "/tmp/a/b/c",  "../..",           "[1] Leftovers in Base" },
   { "/tmp/a",          "/tmp/a/b",    "..",              "[1] Leftovers in Base" },
   { "/tmp",            "/tmp/a/b",    "../..",           "[1] Leftovers in Base" },
   { "/tmp",            "/tmp/a",      "..",              "[1] Leftovers in Base" },
   { "/",               "/tmp",        "..",              "[1] Leftovers in Base" },

   { "/",               "/a",          "..",              "[1] Leftovers in Base" },
   { "/",               "/a/",         "..",              "[1] Leftovers in Base Extra Base Slash" },
   { "/",               "/a/b",        "../..",           "[1] Leftovers in Base" },
   { "/",               "/a/b/",       "../..",           "[1] Leftovers in Base Extra Base Slash" },

   { "/",               "/",           ".",                "Exact same root" },
   { "",                "",            ".",                "Exact same empty" },
   { "/tmp",            "/tmp",        ".",                "Exact same" },
   { "/tmp/foo",        "/tmp",        "foo",              "Simple File Relative To Dir" },
   { "foo",             ".",           "foo",              "File Relative to dot"},
   { "/var/tmpx/a",     "/var/tmp",    "../tmpx/a",        "From Similar Component" },
   { "/var/tmpx/a",     "/var/tmp/",   "../tmpx/a",        "'' Base Only Extra Trailing" },
   { "/var/tmpx/a/",    "/var/tmp/",   "../tmpx/a/",       "'' From+Base Trailing" },
   { "/var/tmpx/a/",    "/var/tmp",    "../tmpx/a/",       "'' From Only Trailing" },

   // From                       Base                         Expect
   { "/var/tmpxyz/foo",          "/var/tmp",                  "../tmpxyz/foo",                 "x" },
   { "/var/tmpxyz/foo",          "/var/tmp/",                 "../tmpxyz/foo",                 "x" },
   { "/test/path/folder/file",   "/test/path_suffix/folder",  "../../path/folder/file",        "x" },
   { "/test/path_suffix/afile",  "/test/path/some/folder",    "../../../path_suffix/afile",    "x" },
   { "/aaa/bbbxxx/foo",          "/aaa/bbb/ccc",              "../../bbbxxx/foo",              "x" },
   { "/aaa/bbbxxx/foo",          "/aaa/bbb/ccc/",             "../../bbbxxx/foo",              "Extra Trailing on Base" },
   { "/aaa/bbb/foo",             "/aaa/bbbxxx/ccc/",          "../../bbb/foo",                 "x" },
   { "/var/log/syslog",          "/var/tmp/../log",           "syslog",                        "x" },
   { "/var/log/syslog",          "/var/tmp",                  "../log/syslog",                 "x" },
   { "/var/tmp/foo",             "/var/tmp",                  "foo",                           "x" },
   { "/var/tmp/foo",             "/var/log/../tmp",           "foo",                           "x" },
   { "/var/log/syslog",          "/var/../var/tmp",           "../log/syslog",                 "x" },

   // NOTE: The following are from chris's tests
   // Small mods:
   //     > we get rid of extra /'s, and never keep them.
   //     > removed tests that attempt to make 'base' a path to a file (must be a dir)
   //
   // From                          Base                            Expect                         Comments
   { "/test/path/folder/file",      "/test/path_suffix/folder",     "../../path/folder/file",      "STR#3441" },
   { "/test/path/folder/file",      "/test/path_suffix/folder/",    "../../path/folder/file",      "STR#3441 Extra Trailing on From" },
   { "/test/path/folder/file",      "/test/path_suffix////folder",  "../../path/folder/file",      "Multiple /'s In Base" },
   { "/test/path/folder/////file",  "/test/path_suffix/folder",     "../../path/folder/file",      "Multiple /'s in From" },
   { "/test/path/folder/////file",  "/test/path_suffix///folder",   "../../path/folder/////file",  "Multiple /'s in Base + From" },
   { "/",                           "/test/path/folder",            "../../..",                    "From is /" },
   { "/test/path/folder/file",      "/",                            "test/path/folder/file",       "Base is /" },
   { "/anydir/test1/subdir/file",   "/anydir/test/subdir",          "../../test1/subdir/file",     "another test for STR#2384" },
   { "/anydir/test/subdir/file",    "/anydir/test1/subdir",         "../../test/subdir/file",      "another test for STR#3441" },
   { "/test/path/",                 "/test/path",                   ".",                           "same path but appended '/'" },
   { "/test/path",                  "/test/path",                   ".",                           "exactly same path" },
   { "/test/path/folder/file",      "/test/path/folder",            "file",                        "same path with file" },
   { 0, 0, 0, 0 }
 };

  for ( int t=0; test[t].from; t++ ) {
      const char *from = test[t].from;
      const char *base = test[t].base;
      printf("\nTEST: %s\n", test[t].comment);
      printf("        Path: '%s'\n", test[t].from);
      printf("        Base: '%s'\n", test[t].base);
      int ret = fl_filename_relative(buf,256,from,base);
      printf("         Got: '%s'\n", buf);
      printf("      Expect: '%s'\n", test[t].expect);
      if ( strcmp(buf, test[t].expect) != 0 )
          { printf("\033[7m*** FAIL! ***\033[0m\n"); }
      if ( ret != 0 )
          { printf("\033[7m*** FAIL NONZERO RETURN ***\033[0m\n"); }
  }
  printf("OK\n");
  return 0;
}
